from jinja2 import Template

postgres_create_table = Template('''
    -- Create table
    CREATE TABLE {{ schema }}.{{ task_id }}_tmp
    AS ({{ sql }});

    -- Drop existing table
    DROP TABLE IF EXISTS {{ schema }}.{{ task_id }};

    -- Replace query
    ALTER TABLE {{ schema }}.{{ task_id }}_tmp RENAME TO {{ task_id }};
''')

postgres_comment_table = Template('''
    -- Table description
    {% if description %}
        COMMENT ON TABLE {{ schema }}.{{ task_id }} IS '{{ description | replace("\'", "\'\'") }}';
    {% endif %}

    -- Field descriptions
    {% if fields %}
        {% for field in fields %}
            {% for field_name, field_description in field.items() %}
                COMMENT ON COLUMN {{ schema }}.{{ task_id }}."{{ field_name }}" IS '{{ field_description | replace("\'", "\'\'") }}';
            {% endfor %}
        {% endfor %}
    {% endif %}
''')



import os

from airflow.contrib.operators.ssh_operator import SSHOperator
from airflow.contrib.hooks.sftp_hook import SFTPHook
from airflow.utils.decorators import apply_defaults

command_template = """
Rscript -e 'library(methods); rmarkdown::render("~/{basename}")'; rm -f ~/{basename}; rm -f ~/{html_out}
"""

class RmdOperator(SSHOperator):
    """
    The RmdOperator executes the R Markdown file. Note that it is up to the Rmd itself
    to handle connecting to the database.
    """
    ui_color = "#75AADB"
    template_fields = SSHOperator.template_fields + ('file_path', )
    
    @apply_defaults
    def __init__(self, file_path, ssh_conn_id = "rserver_default", *args, **kwargs):
        self.file_path = file_path
        self.base_name = os.path.basename(self.file_path)
        self.html_output = self.base_name.replace('.Rmd', '.html')

        command = command_template.format(basename = self.base_name,
                                          html_out = self.html_output)

        super(RmdOperator, self).__init__(ssh_conn_id = ssh_conn_id, command = command,
                                          *args, **kwargs)

    def execute(self, context):
        # Run the command, but first put the file up on the server
        sftp_hook = SFTPHook(ftp_conn_id=self.ssh_conn_id,
                             timeout=self.timeout)
        
        sftp_hook.store_file(self.base_name, self.file_path)

        super(RmdOperator, self).execute(context)



import re

from airflow.operators.postgres_operator import PostgresOperator
from airflow.utils.decorators import apply_defaults

from ..templates.sql_templates import postgres_create_table, postgres_comment_table

def detect_dependencies(query, schema, task_id):
    query = re.sub(re.compile(r'\/\*.*\*\/', re.MULTILINE), "", query)
    query = re.sub("--.*\n", "", query)
    query = re.sub(re.compile(r'[\s]+', re.MULTILINE), " ", query)

    regex = "[^a-z\d_\.]" + schema + "\.([a-z\d_\.]*)"
    query_tables = re.finditer(regex, query)
    ret = list(set(m.group(1) for m in query_tables))

    return ret

###############
## Operators ##
###############

class MaterializedPostgresOperator(PostgresOperator):
    ui_color = "#c37ed5"
    template_fields = PostgresOperator.template_fields + ("schema", "description", "fields", )

    @apply_defaults
    def __init__(
            self,
            task_id,
            sql,
            postgres_conn_id = "postgres_default",
            schema = "views",
            description = None,
            fields = None,
            **kwargs):

        self.schema = schema
        self.description = description
        self.fields = fields

        # Turn the SQL into a CREATE TABLE + document command
        create_sql = postgres_create_table.render(task_id = task_id,
                                                  schema = schema,
                                                  sql = sql)
        doc_sql = postgres_comment_table.render(task_id = task_id,
                                             schema = schema,
                                             description = description,
                                             fields = fields)
        combined_sql = create_sql + "\n" + doc_sql
        
        # Automatically detect dependencies (to be resolved later)
        self.dependencies = detect_dependencies(sql, schema, task_id)

        super(MaterializedPostgresOperator, self).__init__(
            task_id = task_id,
            sql = combined_sql,
            postgres_conn_id = postgres_conn_id,
            **kwargs)



import os
import re
from inflection import underscore

from airflow.models.baseoperator import BaseOperator
from airflow.hooks.postgres_hook import PostgresHook

from airflow.utils.decorators import apply_defaults

from gusty.templates.sql_templates import postgres_comment_table

def clean_columns(df):
    df.columns = df.columns.str.strip()
    df.columns = df.columns.map(lambda x: underscore(x))
    df.columns = df.columns.map(lambda x: re.sub('\'', '', x))
    df.columns = df.columns.map(lambda x: re.sub('\"', '', x))
    df.columns = df.columns.map(lambda x: re.sub('[^0-9a-zA-Z_]+', '_', x))
    df.columns = df.columns.map(lambda x: re.sub('_+', '_', x))
    return df

class PythonToPostgresOperator(BaseOperator):
    """Upload a pandas DataFrame, from a get_data method, to a Postgres database"""
    ui_color = "#fffad8"
    template_fields = BaseOperator.template_fields + ["postgres_conn_id", "schema", "description", "fields"]

    @apply_defaults
    def __init__(
            self,
            postgres_conn_id = "postgres_default",
            schema = "views",
            description = None,
            fields = None,
            **kwargs):

        self.postgres_conn_id = postgres_conn_id
        self.schema = schema
        self.description = description
        self.fields = fields
        super(PythonToPostgresOperator, self).__init__(**kwargs)

    def get_data(self, context):
        raise NotImplementedError("A PythonToPostgresOperator should " +
          "implement a get_data method that returns a pandas DataFrame")

    def execute(self, context):
        # Get dataset from the get_data method
        dataset = self.get_data(context)
        dataset_cleaned = clean_columns(dataset)

        hook = PostgresHook(postgres_conn_id = self.postgres_conn_id)
        engine = hook.get_sqlalchemy_engine()

        # upload it to SQL
        dataset_cleaned.to_sql(name=self.task_id,
                               con=engine,
                               schema=self.schema,
                               if_exists="replace",
                               index=False)
        
        # comment the table
        if self.fields or self.description:
            comment_sql = postgres_comment_table.render(task_id = self.task_id,
                                                        schema = self.schema,
                                                        fields = self.fields,
                                                        description = self.description)

            engine.execute(comment_sql, autocommit = True)



import os

from airflow.utils.decorators import apply_defaults

from .python_to_postgres_operator import PythonToPostgresOperator

class CSVToPostgresOperator(PythonToPostgresOperator):
    """Upload a CSV file from the dags/csv folder to a Postgres connection."""
    ui_color = "#fffad8"
    template_fields = PythonToPostgresOperator.template_fields + ["csv_file"]

    @apply_defaults
    def __init__(
            self,
            csv_file,
            **kwargs):
        self.csv_file = csv_file
        super(CSVToPostgresOperator, self).__init__(**kwargs)

    def get_data(self, context):
        """Read in a CSV file from dags/csv, and upload it to Postgres"""
        csv_dir = os.path.join(os.getenv('AIRFLOW_HOME'), "dags", "csv")
        csv_path = os.path.join(csv_dir, self.csv_file)
    
        if not os.path.exists(csv_path):
            raise FileNotFoundError("CSV %s not found in %s" %
            (self.csv_file, csv_dir))

        import pandas as pd
        csv_data = pd.read_csv(csv_path)
        return csv_data



# Provide plugins through airflow.operators.gusty

from airflow.plugins_manager import AirflowPlugin
from .operators.rmd_operator import RmdOperator
from .operators.jupyter_operator import JupyterOperator
from .operators.materialized_postgres_operator import MaterializedPostgresOperator
from .operators.csv_to_postgres_operator import CSVToPostgresOperator

class GustyPlugin(AirflowPlugin):
  name = 'gusty'
  operators = [RmdOperator, JupyterOperator, MaterializedPostgresOperator,
               CSVToPostgresOperator]



import sys
import re
import os
import yaml
import pkgutil
import itertools

import airflow
from airflow.operators.latest_only_operator import LatestOnlyOperator
from airflow.sensors.external_task_sensor import ExternalTaskSensor

import frontmatter
import nbformat
from inflection import underscore

##########

# Hack: add AIRFLOW_HOME/operators to the Python path, for custom operators
CUSTOM_OPERATORS_DIR = os.path.join(os.environ["AIRFLOW_HOME"], "operators")
sys.path.append(CUSTOM_OPERATORS_DIR)

gusty_path = [os.path.join(os.path.split(__file__)[0], "operators")]

module_paths = [
    ("airflow.operators.", airflow.operators.__path__),
    ("airflow.contrib.operators.", airflow.contrib.operators.__path__),
    ("gusty.operators.", gusty_path),
    ("", [CUSTOM_OPERATORS_DIR])
]

pairs = [[(_.name, m + _.name) for _ in pkgutil.iter_modules(path)]
    for m, path in module_paths]

module_dict = dict(itertools.chain(*pairs))

def __get_operator(operator_name):
    """Given the name as camel case"""
    if not re.match("^[A-Za-z\d]+$", operator_name):
        raise ValueError("Operator name must be an alphanumeric string")

    module_name = module_dict[underscore(operator_name)]
    import_stmt = "from %s import %s" % (module_name, operator_name)

    exec(import_stmt)
    return eval(operator_name)

###############################################
################# File System #################
###############################################

################
## Read Files ##
################

valid_extensions = ('.yml', '.Rmd', '.ipynb')

def get_files(yaml_dir):
    """
    List all file paths in a dag subdirectory
    """
    files = [os.path.join(yaml_dir, file) for file in os.listdir(yaml_dir) if file.endswith(valid_extensions)]
    assert len(files) > 0, "No files with valid extensions found. Valid extensions are " + valid_extensions
    return files

########################
## Reading in yaml specs
########################

def read_yaml_spec(file):
    """
    Reading in yaml specs
    """

    if file.endswith('.ipynb'):
        # Find first yaml cell in jupyter notebook and parse yaml
        nb_cells = nbformat.read(file, as_version=4)['cells']
        yaml_cell = [cell for cell in nb_cells if cell['cell_type'] == 'markdown' and cell['source'].startswith('```yaml')][0]['source']
        yaml_file = yaml.safe_load(yaml_cell.replace('```yaml', "").replace("```", ""))

    else:
        # Read either the frontmatter or the parsed yaml file (using "or" to coalesce them)
        file_parsed = frontmatter.load(file)
        yaml_file = file_parsed.metadata or yaml.load(file_parsed.content, Loader = yaml.FullLoader)

    assert "operator" in yaml_file, "No operator specified in yaml spec " + file

    task_id = os.path.splitext(os.path.basename(file))[0]
    yaml_file["task_id"] = task_id.lower().strip()
    assert yaml_file["task_id"] != "all", "Task name 'all' is not allowed. Please change your task name."

    yaml_file["file_path"] = file

    return yaml_file

def get_yaml_specs(directory, **kwargs):
    yaml_files = get_files(directory)
    assert len(yaml_files) > 0, "No .yml files found."
    specs = list(map(read_yaml_spec, yaml_files))
    return specs

################################################
################# Dependencies #################
################################################

#####################################
## Different kinds of dependencies ##
#####################################

# yaml spec dependencies

def get_yaml_spec_dependencies(spec, task):
    """
    Get yaml spec dependencies from a single yaml spec.
    """
    spec_dependencies = spec.get("dependencies", [])
    task_dependencies = task.dependencies if hasattr(task, "dependencies") else []

    return spec_dependencies + task_dependencies

# External Dependencies

def get_spec_external_dependencies(spec):
    external_dependencies = {}
    external_dependencies["task_id"] = spec["task_id"]
    external_dependencies["external_dependencies"] = spec["external_dependencies"] if "external_dependencies" in spec.keys() else None
    return external_dependencies

############################
## Dependency Aggregation ##
############################

def get_external_dependencies(yaml_specs):
    external_dependencies = [*map(get_spec_external_dependencies, yaml_specs)]
    external_dependencies = [external_dependency for external_dependency in external_dependencies if external_dependency["external_dependencies"] is not None]
    return external_dependencies

######################
## Set Dependencies ##
######################

def set_dependencies(yaml_specs, tasks, latest_only=True, **kwargs):
    dependencies = [(s["task_id"], get_yaml_spec_dependencies(s, tasks[s["task_id"]]))
        for s in yaml_specs]

    external_dependencies = get_external_dependencies(yaml_specs)

    if latest_only:
         latest_only_operator = LatestOnlyOperator(task_id='latest_only', dag=kwargs["dag"])

    #external_dependencies
    external_tasks = {}
    tasks_with_external_dependencies = []

    for task in external_dependencies:
        task_id = task["task_id"]
        tasks_with_external_dependencies.append(task_id)
        external_dependencies = task["external_dependencies"]

        for external_dependency in external_dependencies:
            external_dag, external_task = list(external_dependency.items())[0]
            wait_for_whole_dag = external_task == r"all"
            task_name = "wait_for_DAG_" + external_dag if wait_for_whole_dag else "wait_for_" + external_task

            if task_name not in external_tasks.keys():
                if wait_for_whole_dag:
                    wait_for_task = ExternalTaskSensor(dag = kwargs["dag"],
                                                       task_id = task_name,
                                                       external_dag_id=external_dag,
                                                       external_task_id=None,
                                                       poke_interval=20,
                                                       timeout=60,
                                                       retries=25)
                    external_tasks[task_name] = wait_for_task
                else:
                    wait_for_task = ExternalTaskSensor(dag = kwargs["dag"],
                                                       task_id = task_name,
                                                       external_dag_id=external_dag,
                                                       external_task_id=external_task,
                                                       poke_interval=60,
                                                       timeout=60,
                                                       retries=25)
                    external_tasks[task_name] = wait_for_task

                if latest_only:
                    wait_for_task.set_upstream(latest_only_operator)

            tasks[task_id].set_upstream(external_tasks[task_name])

    # local dependencies
    for task_id, task_depends_on in dependencies:
        valid_dependencies = [d for d in task_depends_on
            if d in tasks.keys() and d != task_id]

        if len(valid_dependencies) == 0 and latest_only and task_id not in tasks_with_external_dependencies:
            tasks[task_id].set_upstream(latest_only_operator)
        else:
            for d in valid_dependencies:
                tasks[task_id].set_upstream(tasks[d])



###############################################
################# Build Tasks #################
###############################################

def build_tasks(yaml_specs, dag):
    task_dict = {}

    for spec in yaml_specs:
        operator = __get_operator(spec["operator"])

        # The spec will have dag added and some keys removed
        args = {k:v for k,v in spec.items()
            if not hasattr(operator, 'template_fields')
                or k in operator.template_fields}
        args["task_id"] = spec["task_id"]
        args["dag"] = dag

        # Some arguments are used only by gusty
        for field in ['operator', 'dependencies', 'external_dependencies']:
            args.pop(field, None)

        task = operator(**args)

        task_dict[spec["task_id"]] = task

    return task_dict

def build_dag(directory, dag, latest_only=True):
    yaml_specs = get_yaml_specs(directory)
    tasks = build_tasks(yaml_specs, dag=dag)
    set_dependencies(yaml_specs, tasks, dag=dag, latest_only=latest_only)

    return tasks
